# zope imports
from zope.interface import implements
from zope.app.event.objectevent import ObjectModifiedEvent

# iqpp imports
from iqpp.rating.interfaces import IRated

class Rated(ObjectModifiedEvent):
    """An Event, which is fired up when a user has rated.
    """
    implements(IRated)
    
    def __init__(self, old_score, score, user, object, context=None):
        """
        """
        self.old_score  = old_score
        self.score  = score
        self.object = object  
        self.user = user
        self.context = context


# zope imports
from zope.interface import implements

# iqpp imports
from iqpp.rating.interfaces import IRatingDefinition

class RatingDefinition:
    """Defines scores and corresponding numerical values.
    """
    
    implements(IRatingDefinition)

    def __init__(self, title, scores, description=None):
        """
        """
        self.title = title
        self.scores = scores
        if description is not None:
            self.description = description

    def isValidScore(self, score):
        """
        """
        if score in [s[0] for s in self.scores]:
            return True
        return False
        
    def getNumericalValue(self, score):
        """
        """
        for s in self.scores:
            if s[0] == score:
                return s[1]
        return 0



# zope imports
from transaction import commit
from zope.interface import directlyProvides

# Testing imports
from Testing import ZopeTestCase
from Products.PloneTestCase import PloneTestCase
from Products.PloneTestCase.layer import PloneSite
from AccessControl.SecurityManagement import newSecurityManager

# Five imports
from Products.Five import zcml

# CMFCore imports
from Products.CMFCore.utils import getToolByName

# iqpp.rating imports
import iqpp.rating
from iqpp.rating.interfaces import IRateable
from iqpp.rating.tests import utils

PRODUCTS = []
PloneTestCase.setupPloneSite(products=PRODUCTS)

class RatingLayer(PloneSite):

    @classmethod
    def setUp(cls):
        app = ZopeTestCase.app()
        portal = app.plone

        zcml.load_config('configure.zcml', iqpp.rating)
        
        setup_tool = getToolByName(portal, 'portal_setup')
        setup_tool.runAllImportStepsFromProfile('profile-iqpp.rating:iqpp.rating')
        
        # login as admin (copied from `loginAsPortalOwner`)
        uf = app.acl_users
        user = uf.getUserById(PloneTestCase.portal_owner).__of__(uf)
        newSecurityManager(None, user)

        commit()
        ZopeTestCase.close(app)

    @classmethod
    def tearDown(cls):
        pass


class RatingMixin:

    def afterSetUp(self):
        """
        """
        self.setRoles(['Manager'])

        # Create default score card
        utils.createDefinition()

        # A rateable document for all tests to play with
        self.portal.invokeFactory("Document", "document")
        directlyProvides(self.portal.document, IRateable)
        
class RatingTestCase(RatingMixin, PloneTestCase.PloneTestCase):
    """Base class for integration tests for the 'iqpp.rating' product.
    """
    layer = RatingLayer

class RatingFunctionalTestCase(RatingMixin, PloneTestCase.FunctionalTestCase):
    """Base class for functional integration tests for the 'iqpp.rating' product.
    """
    layer = RatingLayer



# Zope imports
from AccessControl import Unauthorized
from datetime import datetime
from datetime import timedelta

# zope imports
from zope.interface import directlyProvides
from zope.component import provideUtility, queryUtility

# test imports
from base import RatingTestCase

# iqpp.rating imports
from iqpp.rating.content.rating import Rating    
from iqpp.rating.content.rating import AnonymousRating    

class TestRatingDefinition(RatingTestCase):

    def testRating(self):
        rating = Rating(id=u"score", score=u"1", user=u"john")
        
        self.assertEqual(rating.id, u"score")
        self.assertEqual(rating.user, u"john")
        self.assertEqual(rating.score, u"1")        

class TestRatingDefinitionAnonymousRating(RatingTestCase):

    def afterSetUp(self):
        self.rating = AnonymousRating(id=u"score")

    def testInitialization(self):
        self.assertEqual(self.rating.id, u"score")
        self.assertEqual(self.rating.user, u"anonymous")
        self.assertEqual(self.rating.average_rating, 0.0)        
        self.assertEqual(self.rating.amount_of_ratings, 0)
        self.assertEqual(list(self.rating.ipaddresses.keys()), [])
        
    def testAddValue(self):
        result = self.rating.addValue(2, u"2", u"127.0.0.1")
        self.assertEqual(result, True)
        self.assertEqual(self.rating.id, u"score")
        self.assertEqual(self.rating.user, u"anonymous")
        self.assertEqual(self.rating.average_rating, 2.0)        
        self.assertEqual(list(self.rating.scores.keys()), ["2"])                
        self.assertEqual(self.rating.amount_of_ratings, 1)  
        self.assertEqual(list(self.rating.ipaddresses.keys()), [u"127.0.0.1"])

        result = self.rating.addValue(4, u"4", u"127.0.0.2")
        self.assertEqual(result, True)
        self.assertEqual(self.rating.id, u"score")
        self.assertEqual(self.rating.user, u"anonymous")
        self.assertEqual(self.rating.average_rating, 3.0)
        self.assertEqual(list(self.rating.scores.keys()), [u"2", u"4"])
        self.assertEqual(self.rating.amount_of_ratings, 2)  
        self.assertEqual(list(self.rating.ipaddresses.keys()), [u"127.0.0.1", u"127.0.0.2"])

        result = self.rating.addValue(4, u"4", u"127.0.0.2")
        self.assertEqual(result, False)

    def testAddValueSameIPAfterOneDay(self):
        result = self.rating.addValue(2, "2", "127.0.0.1")
        self.assertEqual(result, True)
        self.assertEqual(self.rating.id, "score")
        self.assertEqual(self.rating.user, "anonymous")
        self.assertEqual(self.rating.average_rating, 2.0)
        self.assertEqual(list(self.rating.scores.keys()), ["2"])
        self.assertEqual(self.rating.amount_of_ratings, 1)
        self.assertEqual(list(self.rating.ipaddresses.keys()), ["127.0.0.1"])
        
        self.rating.ipaddresses["127.0.0.1"] = datetime.now() - timedelta(1)
        
        result = self.rating.addValue(4, "4", "127.0.0.1")
        self.assertEqual(result, True)
        self.assertEqual(self.rating.id, "score")
        self.assertEqual(self.rating.user, "anonymous")
        self.assertEqual(self.rating.average_rating, 3.0)
        self.assertEqual(list(self.rating.scores.keys()), ["2", "4"])
        self.assertEqual(self.rating.amount_of_ratings, 2)  
        self.assertEqual(list(self.rating.ipaddresses.keys()), ["127.0.0.1"])
                                                
def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestRatingDefinition))        
    suite.addTest(makeSuite(TestRatingDefinitionAnonymousRating))
    return suite



from iqpp.rating.config import ENCODING

def stringToUnicode(string):
    """
    """
    if string is None:
        return None

    if isinstance(string, unicode):
        return string

    return unicode(string, ENCODING)


# zope imports
from zope.interface import implements
from zope.schema.fieldproperty import FieldProperty

# Zope imports
from OFS.SimpleItem import SimpleItem

# iqpp.rating imports
from iqpp.rating.interfaces import IRatingOptions

class RatingOptionsStorage(SimpleItem):
    """Stores global rating options.
    """
    implements(IRatingOptions)    
    is_enabled = FieldProperty(IRatingOptions['is_enabled'])
    score_card = FieldProperty(IRatingOptions['score_card'])


# zope imports
from zope.interface import implements
from zope.schema.fieldproperty import FieldProperty
from datetime import datetime
from datetime import timedelta

# Zope imports
from persistent import Persistent
from BTrees.OOBTree import OOBTree

# rating
from iqpp.rating.interfaces import IRating
from iqpp.rating.interfaces import IAnonymousRating

class Rating(Persistent):
    """
    """    
    implements(IRating)

    id = FieldProperty(IRating['id'])
    score = FieldProperty(IRating['score'])
    user = FieldProperty(IRating['user'])
    timestamp = FieldProperty(IRating['timestamp'])

    def __init__(self, id, score, user):
        """
        """
        self.id = id
        self.score = score
        self.user = user
        self.timestamp = datetime.now()
        
class AnonymousRating(Rating):
    """
    """
    implements(IAnonymousRating)

    amount_of_ratings = FieldProperty(IAnonymousRating['amount_of_ratings'])
    average_rating = FieldProperty(IAnonymousRating['average_rating'])    
    
    def __init__(self, id):
        super(AnonymousRating, self).__init__(id, u"0", u"anonymous")
        self.amount_of_ratings = 0
        self.average_rating = 0.0        
        self.ipaddresses = OOBTree()
        self.scores = OOBTree()

    def hasRated(self, ipaddress):
        """
        """
        one_day = timedelta(1)
        if ipaddress in self.ipaddresses and\
           (datetime.now() - self.ipaddresses.get(ipaddress)) < one_day:
           return True
        
        return False
        
    def addValue(self, value, score, ipaddress):
        """
        """
        # If given ipaddress has already rated within a particular time range
        # it isn't allowed to rate again.
        if self.hasRated(ipaddress):
            return False
            
        if self.scores.has_key(score) == False:
            self.scores[score] = 0
        self.scores[score] += 1
            
        self.average_rating = ((self.average_rating * self.amount_of_ratings) + value) / (self.amount_of_ratings + 1)   
        self.amount_of_ratings += 1            
        self.ipaddresses[ipaddress] = datetime.now()
        return True


# Zope imports
from AccessControl import Unauthorized

# zope imports
from zope.interface import directlyProvides
from zope.component import provideUtility, queryUtility

# test imports
from base import RatingTestCase

# iqpp.rating imports
from iqpp.rating.interfaces import IRatingDefinition    
from iqpp.rating.content.definition import RatingDefinition

class TestRatingDefinition(RatingTestCase):

    def testQueryDefinition(self):
        definition = queryUtility(IRatingDefinition, "score")
        
        self.assertEqual(True, definition.isValidScore("3"))
        self.assertEqual(False, definition.isValidScore("42"))
                        
def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestRatingDefinition))
    return suite



# Zope imports
from AccessControl import Unauthorized

# zope imports
from zope.interface import directlyProvides
from zope.component import provideUtility

# test imports
from base import RatingTestCase

# iqpp.rating imports
from iqpp.rating.interfaces import IRateable
from iqpp.rating.interfaces import IRatingManager
from iqpp.rating.interfaces import IRatingDefinition    
from iqpp.rating.content.definition import RatingDefinition

class TestRating(RatingTestCase):

    def testRate(self):
        rm = IRatingManager(self.portal.document)
        
        rm.rate(id=u"score", score=u"1", user=u"john")
        rm.rate(id=u"score", score=u"2", user=u"jane")
        rm.rate(id=u"score", score=u"3", user=u"ben")
                        
        rating = rm.getRating(id="score", user="john")
        self.assertEqual(rating, "1")

        rating = rm.getRating(id="score", user="jane")
        self.assertEqual(rating, "2")

        rating = rm.getRating(id="score", user="ben")
        self.assertEqual(rating, "3")

        ratings = rm.getRatings(id="score")
        self.assertEqual(len(ratings), 3)

        ratings = rm.getAmountOfRatings(id="score")
        self.assertEqual(ratings, 3)
        
        average = rm.getAverageRating(id="score")
        self.assertEqual(average, 2.0)

    def testNoRatings(self):
        rm = IRatingManager(self.portal.document)
        average = rm.getAverageRating(id="score")
        self.assertEqual(average, -1)
        
    def testWrongValue(self):
        rm = IRatingManager(self.portal.document)
        self.assertRaises(ValueError, rm.rate, id="score", score="42", user="john")

    def testWrongDefinition(self):
        rm = IRatingManager(self.portal.document)
        self.assertRaises(ValueError, rm.rate, id="wrong", score="1", user="john")

    def testChangeRating(self):
        rm = IRatingManager(self.portal.document)
        rm.rate(id=u"score", score=u"1", user=u"john")

        rating = rm.getRating(id=u"score", user=u"john")
        self.assertEqual(rating, "1")
        
        rm.rate(id=u"score", score=u"2", user=u"john")
        self.assertEqual(rating, "1")        

class TestRatingAsAnonymous(RatingTestCase):

    def testRate_1(self):
        """
        """
        rm = IRatingManager(self.portal.document)
        self.logout()
        
        rm.rate(id="score", user=None, score="1", ipaddress="127.0.0.1")
        rm.rate(id="score", user=None, score="2", ipaddress="127.0.0.2")
        rm.rate(id="score", user=None, score="3", ipaddress="127.0.0.3")

        self.login("test_user_1_")
        rating = rm.getRating(id="score")
        self.assertEqual(rating, 2.0)

        ratings = rm.getRatings(id="score")
        self.assertEqual(len(ratings), 1)

        ratings = rm.getAmountOfRatings(id="score")
        self.assertEqual(ratings, 3)
        
        average = rm.getAverageRating(id="score")
        self.assertEqual(average, 2.0)

    def testRate_2(self):
        rm = IRatingManager(self.portal.document)
        self.assertRaises(Unauthorized, rm.rate, id="score", score="1", user="john")
        
def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestRating))        
    # suite.addTest(makeSuite(TestRatingAsAnonymous))
    return suite



def importVarious(context):
    """Import various settings.
    """
    # Set default permissions
    portal = context.getSite()
    portal.manage_permission('iqpp.rating: rate',    ["Anonymous", "Member"], 1)
    portal.manage_permission('iqpp.rating: details', ["Anonymous", "Member"], 1)



# zope imports
from zope.app.annotation.interfaces import IAttributeAnnotatable
from zope.i18nmessageid import MessageFactory
from zope.interface import Interface, Attribute
from zope.viewlet.interfaces import IViewletManager
from zope import schema
_ = MessageFactory("iqpp.rating")

class IRatingViewletManager(IViewletManager):
    """Viewlet manager for rating.
    """

class IRateable(IAttributeAnnotatable):
    """A marker interface marking an object as being ratable.
    """

class IRatingManager(Interface):
    """An interface which provides methods for rating arbitrary objects.
    """        
    def deleteRating(id, user):
        """Deletes the rating for the definition id and user. Returns True if 
        the rating was deleted, otherwise returns False.
        """
            
    def getAmountOfRatings(id):
        """Returns the total amount of ratings for the definition with given 
        id.
        """

    def getAverageRating(id):
        """Returns the average rating value for the definition with given id.
        """

    def getRatings(id):
        """Returns all ratings for the definition with the given id. Returns 
        a list of strings (member ratings) and a float (anonymous).
        """

    def getRating(id, user):
        """Returns a rating for the definition and user. Returns the ratings 
        for anoymous users if given user is None. Returns None if there is no
        rating (e.g. in order to differ zero from no rating).
        """

    def getRatingDefinition(id):
        """Returns the rating definition by given id.
        """
        
    def getRatingId():
        """Returns the rating id for an object. If the object has only one 
        rating, in this the passing of an rating id for every single request
        could be omitted.
        """
                
    def getScores(id):
        """Returns a dict of score : amount, whereas the keys are the scores 
        of the definition and the values are the amount of how often the score 
        was rated. If the score wasn't rated at all, it isn't within the dict.
        """

    def hasRated(id, user):    
        """Returns True if the given user has already rated for the given id.
        """

    def rate(user, id, score, ipaddress):
        """Adds a new rating with given informations.

        Parameters:
            id:    The id of the rating definition.
            score: The id of the score (not the value)        
        """

    def rateWithConfirmation(id, value, user):
        """Checks confirmation and calls the rate method.
        """

    def setRating(id, value, user):
        """Sets a value for a specific id and user. 
        """    
        
class IRating(Interface):
    """A single rating for a definition and user.
    """
    # TODO: Check this, as the id is the id of the rating definition atm. Should 
    # this rather stored in a definition field and additionally a unique id in the 
    # id field?
    id = schema.TextLine(title=_(u"Id"),
         description=_(u"The id of the rating"),
         required=True,   
    )

    score = schema.TextLine(title=_(u"Score"),
         description=_(u"The score which has been rated."),
         required=True,   
    )

    user = schema.TextLine(title=_(u"User"),
         description=_(u"The user who has rated."),
         required=True,   
    )

    timestamp = schema.Datetime(title=_(u"Timestamp"),
         description=_(u"When has been rated"),
         required=True,   
    )

class IAnonymousRating(IRating):
    """Total rating of anonymous users for a specific definition.
    """
    amount_of_ratings = schema.Int(title=_(u"Amount of ratings"),
         description=_(u"The total amount of ratings"),
         required=True,   
         default=0,
    )

    average_rating = schema.Float(title=_(u"Average rating"),
         description=_(u"The average rating"),
         required=True,
         default=0.0,  
    )

    # TODO: Make these as schema fields to give an admin the possibility to 
    # change ratings in future.
    ipaddresses = Attribute("The ip address with which was been rated.")
    scores = Attribute("The scores which have been rated.")
        
    def hasRated(ipaddress):
        """Returns True if the given ip address has already rated.
        """
            
    def addValue(self, value, ipaddress):
        """Adds a value and calculates the new value (which is the average 
        rating over all added ratings. Takes also care of the amount of
        ratings.)
        """

class IRatingDefinition(Interface):
    """A definition for a rating.
    """
    
    def isValidScore(score):
        """Returns true if the given score is valid for this definition.
        """
        
    def getNumericalValue(score):
        """Returns the numerical value for a given score.
        """

class IRated(Interface):
    """A marker interface for the event: user has rated.
    """
        
class IRatingOptions(Interface):
    """Provides a schema for rating options.
    """
    is_enabled = schema.Bool(title=_(u"Enabled"),
         description=_(u"Is rating enabled?"),
         default=True,
    )
    
    score_card = schema.Text(
        title=_(u'The Score Card'),
        description=_(u"The score card. One title:value pair per line."),
        default=u'',
        required=True)
    
    


# zope imports
from zope.interface import implements
from zope.component import adapts
from zope.annotation.interfaces import IAnnotations

# iqpp.rating imports
from iqpp.rating.interfaces import IRatingOptions
from iqpp.rating.interfaces import IRateable
from iqpp.rating.content.options import RatingOptionsStorage

KEY = "iqpp.rating.options"
                
class RatingOptions(object):
    """An adapter which provides IRatingOptions for ratable objects.    
    """
        
    # Global options are stored in RatingOptionsStorage, local are stored as 
    # annotations within the context. We try first to get a local option. If 
    # there is none, we take the global opions.
    
    implements(IRatingOptions)
    adapts(IRateable)

    def __init__(self, context):
        """
        """
        self.context = context
        self.annotations = IAnnotations(context).get(KEY, None)
        if self.annotations is None:
            self.annotations = IAnnotations(context)[KEY] =\
                {'options' : RatingOptionsStorage(), 'overrides' : []}

        self.options = self.annotations['options']
        self.overrides = self.annotations['overrides']

    def getGlobalOption(self, name):
        """3rd-party products may override this method to take their global 
        options from somewhere else, e.g. a configlet. See iqpp.plone.rating 
        for more.
        """
        return IRatingOptions.get(name).default

    def getLocalOption(self, name):
        """3rd-party products may override this method to take their local 
        options from somewhere else, e.g. a configlet. See iqpp.plone.rating 
        for more.
        """
        return getattr(self.options, name)        
        
    def named_get(self, name):
        if name in self.overrides:
            return self.getLocalOption(name)
        else:
            return self.getGlobalOption(name)

    def named_set(self, name, value):
        if name not in self.overrides:
            self.overrides.append(name)
        setattr(self.options, name, value)
        
    @apply
    def is_enabled():
        def get(self): return self.named_get('is_enabled')
        def set(self, value): self.named_set('is_enabled', value)
        return property(get, set)
        
    @apply
    def score_card():
        def get(self): return self.named_get('score_card')
        def set(self, value): self.named_set('score_card', value)
        return property(get, set)


# Zope imports
from DateTime import DateTime

# zope imports
from zope.interface import implements
from zope.component import adapts, queryUtility
from zope.app.annotation.interfaces import IAnnotations
from zope.event import notify as znotify
from BTrees.OOBTree import OOBTree


# CMFCore imports
from Products.CMFCore.utils import getToolByName

# iqpp.rating imports
from iqpp.rating.content.rating import Rating
from iqpp.rating.content.rating import AnonymousRating
from iqpp.rating.events import Rated
from iqpp.rating.interfaces import IRateable
from iqpp.rating.interfaces import IRatingManager
from iqpp.rating.interfaces import IRatingDefinition

KEY1 = "iqpp.rating.members"
KEY2 = "iqpp.rating.anonymous"
KEY3 = "iqpp.rating.daily"

class RatingManager(object):
    """Adapter to manage ratings for arbitrary objects.
    """
    implements(IRatingManager)
    adapts(IRateable)
    
    def __init__(self, context):
        """
        """
        self.context = context    
        annotations = IAnnotations(context)
        
        self.member_ratings = annotations.get(KEY1)
        if self.member_ratings is None:
            self.member_ratings = annotations[KEY1] = OOBTree()

        self.anonymous_ratings = annotations.get(KEY2)
        if self.anonymous_ratings is None:
            self.anonymous_ratings = annotations[KEY2] = OOBTree()

        self.daily_ratings = annotations.get(KEY3)
        if self.daily_ratings is None:
            self.daily_ratings = annotations[KEY3] = {
                "last_rating" : DateTime().earliestTime()-1,
                "amount_of_ratings" : 0,
            }
        
    def deleteRating(self, id, user):
        """
        """
        try:
            del self.member_ratings[id][user]
        except KeyError:
            return False
        
        return True    
        
    def getAmountOfRatings(self, id=None):
        """
        """
        if id is None:
            id = self.getRatingId()
        
        amount_member_ratings = 0
        amount_anonymous_ratings = 0
        
        rating_definition = self.getRatingDefinition(id)
        
        member_ratings = self.member_ratings.get(id)
        if member_ratings is not None:            
            amount_member_ratings = len(member_ratings)
        
        anonymous_ratings = self.anonymous_ratings.get(id)
        if anonymous_ratings is not None:                
            amount_anonymous_ratings = anonymous_ratings.amount_of_ratings

        return amount_member_ratings + amount_anonymous_ratings

    def getAverageRating(self, id=None):
        """
        """
        if id is None:
            id = self.getRatingId()
        
        definition = self.getRatingDefinition(id)
                
        total_value  = 0
        total_amount = 0

        # member ratings
        ratings = self.member_ratings.get(id, {}).values()
        for rating in ratings:
            total_value  += definition.getNumericalValue(rating.score)
            total_amount += 1

        # anonymous ratings (weighted)
        rating = self.anonymous_ratings.get(id)
        if rating is not None:
            total_value  += (rating.average_rating * rating.amount_of_ratings)
            total_amount += rating.amount_of_ratings
            
        try:
            return "%.1f" % (total_value/total_amount)
        except ZeroDivisionError:
            return -1
            
    def getRatingDefinition(self, id):
        """
        """
        rating_definition = queryUtility(IRatingDefinition, id)
        if rating_definition is None:
            raise ValueError('No rating definition named %s found.' % id)

        return rating_definition

    def getRatings(self, id=None):
        """
        """
        if id is None:
            id = self.getRatingId()
        
        rating_definition = self.getRatingDefinition(id)
        ratings = list(self.member_ratings.get(id, {}).values())

        if self.anonymous_ratings.get(id) is not None:
            ratings.append(self.anonymous_ratings.get(id))
            
        return ratings
    
    def getRating(self, id=None, user=None):
        """
        """        
        if id is None:
            id = self.getRatingId()
        
        rating_definition = self.getRatingDefinition(id)
        if user is None:
            try:
                return self.anonymous_ratings[id].score
            except KeyError:
                return None
        else:    
            try:
                return self.member_ratings[id][user].score
            except KeyError:
                return None

    def getRatingId(self):
        """
        """
        try:
            return self.context.getRatingId()
        except AttributeError:
            return None
        
    def getScores(self, id=None):
        """
        """
        if id is None:
            id = self.getRatingId()
        
        scores = {}        
        
        member_ratings = self.member_ratings.get(id, {}).values()        
        for rating in member_ratings:
            if scores.has_key(rating.score) == False:
                scores[rating.score] = 0
            scores[rating.score] += 1    

        anonymous_ratings = self.anonymous_ratings.get(id)
        if anonymous_ratings is not None:        
            for score in anonymous_ratings.scores.keys():
                if scores.has_key(score) == False:
                    scores[score] = 0
                scores[score] += anonymous_ratings.scores.get(score)
        
        return scores

    def hasRated(self, id, user):
        """
        """
        if self.getRating(id, user) is not None:
            return True

        return False
        
    def rate(self, user, id, score, ipaddress=None, context=None, notify=True):
        """
        """
        rating_definition = self.getRatingDefinition(id)
        if not rating_definition.isValidScore(score):
            raise ValueError('Invalid rating score %s for %s.' % (score, id))
        
        # Anonymous user
        if user is None:
            if id not in self.anonymous_ratings:
                self.anonymous_ratings[id] = AnonymousRating(id)

            value = rating_definition.getNumericalValue(score)
            rating = self.anonymous_ratings[id]
            old_score = rating.score
            result = rating.addValue(value, score, ipaddress)
                        
            if result == True:            
                result = "rating-added"
                self._calcDailyRating()
            else:
                result = "already-rated"
        # Members
        else:                                
            if id not in self.member_ratings:
                self.member_ratings[id] = OOBTree()
            
            rating = self.member_ratings[id].get(user)
            if rating is None:
                old_score = 0
                result = "rating-added"
                self._calcDailyRating()
            else:
                old_score = rating.score
                result = "rating-changed"
            self.member_ratings[id][user] = Rating(id, score, user)

            # sent event
            if notify == True:
                if context is None: context = self.context
                znotify(Rated(old_score, score, user, self.context, context))

        self.context.reindexObject()
        return result
        
    def rateWithConfirmation(self):
        """
        """    
        # not implemented yet
     
    def setRating(self, score, id, user):
        """
        """
        # not implemented yet

    def _calcDailyRating(self):
        """
        """
        last_rating = self.daily_ratings["last_rating"]
        today = DateTime().earliestTime()
        if today - last_rating >= 1:
            self.daily_ratings["last_rating"] = DateTime().earliestTime()
            self.daily_ratings["amount_of_ratings"] = 0
            
        self.daily_ratings["last_rating"] = today
        self.daily_ratings["amount_of_ratings"] += 1



# Zope imports
from DateTime import DateTime
from BTrees.OOBTree import OOBTree

# test imports
from base import RatingTestCase

# iqpp.rating imports
from iqpp.rating.interfaces import IRateable
from iqpp.rating.interfaces import IRatingManager
from iqpp.rating.interfaces import IRatingDefinition    
from iqpp.rating.content.definition import RatingDefinition

class TestRatingManager(RatingTestCase):
    """
    """                        
    def testInitialization(self):
        """
        """
        rm = IRatingManager(self.portal.document)        
        daily = rm.daily_ratings

        self.failUnless(len(rm.member_ratings) == 0)
        self.failUnless(isinstance(rm.member_ratings, OOBTree))
        
        self.failUnless(len(rm.anonymous_ratings) == 0)
        self.failUnless(isinstance(rm.anonymous_ratings, OOBTree))
        
        # daily ratings
        self.assertEqual(daily["last_rating"], DateTime().earliestTime()-1)
        self.assertEqual(daily["amount_of_ratings"], 0)

    def testRate_1(self):
        """Test rating generally.
        """    
        rm = IRatingManager(self.portal.document)        
        rm.rate(id=u"score", score=u"1", user=u"john")
        rm.rate(id=u"score", score=u"2", user=u"jane")
        rm.rate(id=u"score", score=u"3", user=u"ben")

        rm = IRatingManager(self.portal.document)        
        daily = rm.daily_ratings
        self.assertEqual(daily["last_rating"], DateTime().earliestTime())
        self.assertEqual(daily["amount_of_ratings"], 3)

    def testRate_2(self):
        """Test daily ratings
        """    
        rm = IRatingManager(self.portal.document)
        rm.rate(id=u"score", score=u"1", user=u"john")
        rm.rate(id=u"score", score=u"2", user=u"jane")
        rm.rate(id=u"score", score=u"3", user=u"ben")

        rm = IRatingManager(self.portal.document)
        daily = rm.daily_ratings

        # Move ratings back on day ...
        yesterday = DateTime()-1
        daily["last_rating"] = yesterday.earliestTime()

        # ... so that they here get deleted.
        rm.rate(id=u"score", score=u"1", user=u"jonny")                
        self.assertEqual(daily["last_rating"], DateTime().earliestTime())
        self.assertEqual(daily["amount_of_ratings"], 1)
        
    def testGetRatings_1(self):
        """Just member ratings
        """
        rm = IRatingManager(self.portal.document)        
        rm.rate(id=u"score", score=u"1", user=u"john")
        rm.rate(id=u"score", score=u"2", user=u"jane")
        rm.rate(id=u"score", score=u"3", user=u"ben")
    
        ratings = rm.getRatings(id="score")
        self.assertEqual(len(ratings), 3)
                
        scores = [r.score for r in ratings]
    
        self.failUnless(u"1" in scores)
        self.failUnless(u"2" in scores)        
        self.failUnless(u"3" in scores)
            
    def testGetRating(self):
        """
        """        
        rm = IRatingManager(self.portal.document)        
        rm.rate(id=u"score", score=u"1", user=u"john")
        rm.rate(id=u"score", score=u"2", user=u"jane")
        rm.rate(id=u"score", score=u"3", user=u"ben")
                
        rating = rm.getRating(id="score", user="john")
        self.assertEqual(rating, "1")
    
        rating = rm.getRating(id="score", user="jane")
        self.assertEqual(rating, "2")
    
        rating = rm.getRating(id="score", user="ben")
        self.assertEqual(rating, "3")
    
    def testGetRatingId(self):
        """
        """
    
    def testGetAmountOfRatings(self):
        """
        """
    
    def testGetAverageRating(self):
        """
        """
    
    def testGetScores(self):
        """
        """
    
    def testHasRated(self):
        """
        """
        
    def testDeleteRating(self):
        """
        """
    
    def testSetRating(self):
        """
        """
        # not implemented yet    
        
    def testRateWithConfirmation(self):
        """
        """    
        # not implemented yet
        
    
def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestRatingManager))        
    return suite



# zope imports 
from zope.component import provideUtility
from zope.interface import implements

# iqpp.rating imports
from iqpp.rating.interfaces import IRateable
from iqpp.rating.interfaces import IRatingDefinition
from iqpp.rating.content.definition import RatingDefinition

def createDefinition():
    """
    """
    sc = [("1", 1.0), ("2", 2.0), ("3", 3.0), ("4", 4.0),  ("5", 5.0),  ("6", 6.0)]
    score = RatingDefinition("Score", sc, "The Score for a Paper")
    provideUtility(score, IRatingDefinition, name='score')


# zope imports
from zope.component.exceptions import ComponentLookupError

# CMFPlone imports
from Products.CMFPlone.CatalogTool import registerIndexableAttribute

# iqpp.rating imports
from iqpp.rating.interfaces import IRatingManager

def average_rating(object, portal, **kwargs):
    try:
        return IRatingManager(object).getAverageRating("plone")            
    except (ComponentLookupError, TypeError, ValueError):
        raise AttributeError

def amount_of_ratings(object, portal, **kwargs):
    try:
        return IRatingManager(object).getAmountOfRatings("plone")            
    except (ComponentLookupError, TypeError, ValueError):
        raise AttributeError

def last_rating(object, portal, **kwargs):
    try:
        return IRatingManager(object).daily_ratings["last_rating"]
    except (ComponentLookupError, TypeError, ValueError):
        raise AttributeError

def daily_amount_of_ratings(object, portal, **kwargs):
    try:
        return IRatingManager(object).daily_ratings["amount_of_ratings"]
    except (ComponentLookupError, TypeError, ValueError):
        raise AttributeError
        
registerIndexableAttribute('average_rating', average_rating)
registerIndexableAttribute('amount_of_ratings', amount_of_ratings)            
registerIndexableAttribute('last_rating', last_rating)
registerIndexableAttribute('daily_amount_of_ratings', daily_amount_of_ratings)            